"""
The unstable package contains experimental functions and classes currently
deemed unstable. The contents of this package is subject to change without
notice, even in minor or patch releases.

Whenever you import anything from the unstable package, you should make sure to
run a type checker such as mypy to help catch these changes.
"""
